/*!
 *	This library is dependant on libfmod, liblua, liblualib & libLuaBinder. 
 *	If you do not want lua function bindings within your code, add the option 
 *	-DNO_LUA in the COMPILER options for the library makefile. This will remove 
 *	the dependancies with the lua scripting language. 
 */

#ifndef __LIB_SOUND___H__
#define __LIB_SOUND___H__

#ifdef __cplusplus
extern "C" {
#endif

	#include <luaBinder.h>

	#ifdef WIN32
		#pragma comment(lib,"fmod.lib")
		#pragma comment(lib,"LibSound.lib")
	#endif

	/*!	
	 *	\brief	Call this function to initialise all of the sound stuff.
	 */
	#ifdef NO_LUA
	void LibSoundInit();
	#else
	void LibSoundInit(lua_State* L);
	#endif
	
	/*!
	 *	\brief	Call this function to kill off the sound stuff
	 */
	void LibSoundCleanup();
	
	/*!
	 *	\brief	Call this function in your idle loop to update all of the sound stuff.
	 */	
	void LibSoundIdle(float dt);

	/*--------------------------------------------------------------------------------*\
	 *
	 *
	 *						---===###	SPECTRUM DATA	###===---
	 *
	 *
	 *		This is only a single function that returns a pointer to 512 floating point
	 *	values. This basically gives you the spectrum data for the current sound.
	 *
	\*--------------------------------------------------------------------------------*/


	/*!
	 *	\brief	This function returns a pointer to 512 float values, aka the spectrum
	 *			data. Have a look at the 08_waveforms example file for more details
	 *	\return	A pointer to 512 float values.
	 */
	float* GetSpectrumData();

#ifdef __cplusplus
}
#endif


#endif

