/*! 
 *
 *
 */
 
#ifndef __LUA_FUNCTION_BINDERS____H_
#define __LUA_FUNCTION_BINDERS____H_

/* automatically link under WIN32 */
#ifdef WIN32
	#pragma comment(lib,"lua.lib")
	#pragma comment(lib,"lualib.lib")
	#pragma comment(lib,"LibLuaBinder.lib")
#endif

 
#ifdef __cplusplus
extern "C" {
#endif
 
	#include <lua.h>
	#include <lauxlib.h>
	#include <lualib.h>
	
	/*!
	 *	Call this to initialise the function binder. Just pass in the initialised
	 *	LUA state. 
	 */
	extern void InitBinder(lua_State* L);
 
	/*!
	 *	\brief	The LUA scripting language allows you to add your own C functions into the
	 *			language. The way LUA does this is probably a bit too hectic for a first
	 *			year. This function therefore is intended to simplify the process. The code
	 *			below may scare you to death, however you are best looking at the examples
	 *			of how to use it rather than trying to understand what I'm doing!!!!!!
	 *	\note	As an example of how to bind funcs to LUA, have a look at these...
	 *
	 *			\code
	 *			void helloWorld() {
	 *				printf("hello world!\n");
	 *			}
	 *			LUA_FUNC0( void, helloWorld );				//	zero arguments, returns void...
	 *
	 *			int add2nums(int a,int b) {
	 *				return a+b;
	 *			}
	 *			LUA_FUNC2( int, add2nums, int, int );		//	two arguments, returns int
	 *
	 *			void print_string(char text[]) {
	 *				printf("string = \"%s\"\n",text);
	 *			}
	 *			LUA_FUNC1( void, print_string, string );	//	1 string argument, (NOTE USE OF 'string' not char* !!) returns void
	 *
	 *
	 *			// in your Init function you should register them, ie :
	 *			REGISTER_FUNC( helloWorld );
	 *			REGISTER_FUNC( add2nums );
	 *			REGISTER_FUNC( print_string );
	 *			\endcode
	 *
	 *			Currently the only supprted argument/return types are :
	 *			\code
	 *			char
	 *			short
	 *			int
	 *			long
	 *			float
	 *			double
	 *			string		->	note the actual C-Type will be char* or char[]. When specifying this in the macro, use string!
	 *			void		->	return type only!
	 *			bool		->	return type only!
	 *			\endcode
	 *
	 *			Array and custom structure types are not supported!
	 *
	 *			To register a function simply use the LUA_FUNC# macro after the function definition.
	 *			The # indicates the number of arguments that the function takes, ie, LUA_FUNC3 is for
	 *			a function that takes 3 arguments.
	 *
	 *			The arguments are Return-Type , function-name, [ arg-type0 , arg-type1 ... ]
	 *
	 *			Having done that you should then register the functions in your Init() function using the
	 *			REGISTER_FUNC( func-name ) macro.
	 *
	 */
	#define LUA_FUNC0(RET,X)								int func_##X (lua_State* L) { if(!_ProcessArgs(L,0)) return 0; ret_##RET (L, X() ); return _NumReturnArgs(); }
	#define LUA_FUNC1(RET,X,T0)								int func_##X (lua_State* L) { if(!_ProcessArgs(L,1)) return 0; ret_##RET (L, X(Arg_##T0 (0))); return _NumReturnArgs(); }
	#define LUA_FUNC2(RET,X,T0,T1)							int func_##X (lua_State* L) { if(!_ProcessArgs(L,2)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1))); return _NumReturnArgs(); }
	#define LUA_FUNC3(RET,X,T0,T1,T2)						int func_##X (lua_State* L) { if(!_ProcessArgs(L,3)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2))); return _NumReturnArgs(); }
	#define LUA_FUNC4(RET,X,T0,T1,T2,T3)					int func_##X (lua_State* L) { if(!_ProcessArgs(L,4)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2),Arg_##T3 (3))); return _NumReturnArgs(); }
	#define LUA_FUNC5(RET,X,T0,T1,T2,T3,T4)					int func_##X (lua_State* L) { if(!_ProcessArgs(L,5)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2),Arg_##T3 (3),Arg_##T4 (4))); return _NumReturnArgs(); }
	#define LUA_FUNC6(RET,X,T0,T1,T2,T3,T4,T5)				int func_##X (lua_State* L) { if(!_ProcessArgs(L,6)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2),Arg_##T3 (3),Arg_##T4 (4),Arg_##T5 (5))); return _NumReturnArgs(); }
	#define LUA_FUNC7(RET,X,T0,T1,T2,T3,T4,T5,T6)			int func_##X (lua_State* L) { if(!_ProcessArgs(L,7)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2),Arg_##T3 (3),Arg_##T4 (4),Arg_##T5 (5),Arg_##T6 (6))); return _NumReturnArgs(); }
	#define LUA_FUNC8(RET,X,T0,T1,T2,T3,T4,T5,T6,T7)		int func_##X (lua_State* L) { if(!_ProcessArgs(L,8)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2),Arg_##T3 (3),Arg_##T4 (4),Arg_##T5 (5),Arg_##T6 (6),Arg_##T7 (7))); return _NumReturnArgs(); }
	#define LUA_FUNC9(RET,X,T0,T1,T2,T3,T4,T5,T6,T7,T8)		int func_##X (lua_State* L) { if(!_ProcessArgs(L,9)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2),Arg_##T3 (3),Arg_##T4 (4),Arg_##T5 (5),Arg_##T6 (6),Arg_##T7 (7),Arg_##T8 (8))); return _NumReturnArgs(); }
	#define LUA_FUNC10(RET,X,T0,T1,T2,T3,T4,T5,T6,T7,T8,T9)	int func_##X (lua_State* L) { if(!_ProcessArgs(L,10)) return 0; ret_##RET (L, X(Arg_##T0 (0),Arg_##T1 (1),Arg_##T2 (2),Arg_##T3 (3),Arg_##T4 (4),Arg_##T5 (5),Arg_##T6 (6),Arg_##T7 (7),Arg_##T8 (8),Arg_##T9 (9))); return _NumReturnArgs(); }

	#define REGISTER_FUNC(X) _Register( #X , (void*) func_##X );
	#define NEW_REGISTER_FUNC(X) {#X,func_##X}
	

	/*--------------------------------------------------------------------------------*\
	 *
	 *
	 *		INTERNAL STUFF		-	DO NOT LOOK !!!!!!!!!!!!!!!  ;)
	 *
	 *
	 *	FYI:	this stuff needs to be defined in this header because it is actually used
	 *			within the LUA_FUNC# and REGISTER_FUNC macros defined above. If they
	 *			were not defined then there would be lots of compiler errors when you
	 *			used the macros.
	 *
	 *			As an example of how the macros work, consider the following :
	 *
	 *			#define EXAMPLE( X ) Arg_ ## X
	 *
	 *			EXAMPLE(int)	expands to 		Arg_int
	 *			EXAMPLE(char)	expands to 		Arg_char
	 *
	 *			The '##' concatenates Arg_ and X together. Using a previous example,
	 *
	 *			LUA_FUNC2( int, add2nums, int, int );
	 *
	 *			becomes -> (with added comments and formatting... )
	 *
	 *			// a new function is created which takes a LUA state
	 *			//
	 *			int func_add2nums(lua_State* L)
	 *			{
	 *				// an internal function is called to process all arguments passed
	 *				// to this function. We request 2 arguments to be parsed. If this fails
	 *				// then the user specified the incorrect number of args
	 *				//
	 *				if( !_ProcessArgs(L,2) )
	 *					return 0;
	 *
	 *				// Arg_int(0) & Arg_int(1) read back the arguments as integers.
	 *				// This is passed into the actual C function call add2nums
	 *				// the return argument from that is passed into a function to interpret
	 *				// it correctly, ie ret_int. This pushes the value onto the LUA stack
	 *				//
	 *				ret_int( add2nums(Arg_int(0),Arg_int(1)) );
	 *
	 *				// This returns the number of return arguments recieved, either 0 or 1
	 *				//
	 *				return _NumReturnArgs();
	 *			}
	 *
	\*--------------------------------------------------------------------------------*/

	void _RegisterLib(char*,luaL_reg[]);
	void _Register(char*,void*);
	int _NumReturnArgs();
	int _ProcessArgs(lua_State*,int);

	/*
	 *	These functions are a cunning hack that uses a macro to determine the arg type
	 */

	/* Ingore me. Used internally.... */
	char   Arg_char(int id);
	int    Arg_int(int id);
	long   Arg_long(int id);
	short  Arg_short(int id);
	float  Arg_float(int id);
	double Arg_double(int id);
	char*  Arg_string(int id);


	#define ret_void(L,X) X
	void    ret_bool(lua_State*,int);
	void    ret_char(lua_State*,char);
	void    ret_int(lua_State*,int);
	void    ret_short(lua_State*,short);
	void    ret_long(lua_State*,long);
	void    ret_float(lua_State*,float);
	void    ret_double(lua_State*,double);
	void    ret_string(lua_State*,char*);
	

#ifdef __cplusplus
}
#endif
 	
	
#endif
