/*!
 *	This library is dependant on libfmod, liblua, liblualib & libLuaBinder. 
 *	If you do not want lua function bindings within your code, add the option 
 *	-DNO_LUA in the COMPILER options for the library makefile. This will remove 
 *	the dependancies with the lua scripting language. 
 */

#ifndef __MP3_PLAYBACK__H__
#define __MP3_PLAYBACK__H__

#ifdef __cplusplus
extern "C" {
#endif

	/*--------------------------------------------------------------------------------*\
	 *
	 *
	 *						---===###	MP3 PLAYBACK	###===---
	 *
	 *
	 *		These functions should allow you to have a nice easy way of playing Mp3's,
	 *	ogg files and other streamed media. Infact, this should work with pretty much
	 *	any media you can come up with....
	 *
	\*--------------------------------------------------------------------------------*/

	/*!
	 *	\brief	This function will load and start playback of the specified mp3 file.
	 *	\param	filename	-	the name of the mp3 to play.
	 *	\return	1 if OK, 0 if failed.
	 */
	int Mp3Play(const char filename[]);

	/*!
	 *	\brief	This function stops playback of the MP3 file.
	 *	\return	1 if OK, 0 if failed.
	 */
	int Mp3Stop();

	/*!
	 *	\brief	This function can pause/unpause playback of the mp3.
	 *	\param	pause	-	1 to pause, 0 to unpause.
	 *	\return	1 if OK, 0 if failed.
	 */
	int Mp3Pause(char);

	/*!
	 *	\brief	This function sets the volume level of the Mp3 between 0 (silent) and 1 (full)
	 *	\param	volume	-	the volume level of the mp3 playback.
	 *	\return	1 if OK, 0 if failed.
	 */
	int Mp3Volume(float);

	/*!
	 *	\brief	This function returns 1 if the mp3 is playing, 0 otherwise. This can be
	 *			called in your Idle function to see if the mp3 has finished and if you
	 *			want to start playback of a new mp3.
	 *	\return	1 if playing, 0 otherwise.
	 */
	int IsMp3Playing();

	/*!
	 *	\brief	Updates the mp3 player...
	 */
	void Mp3Update(float dt);

	
#ifdef __cplusplus
}
#endif

#endif
