//---------------------------------------------------------------------------------------
// pinocchio.cpp written and implimented by Paul Lumley 17/02/05
// this is the main source file for the implimentation of PINOCCHIO V1.1
//
// PINOCCHIO is a mayaAPI plugin which controls specifically rigged characters
// via the use of a standard joypad, the plugin has been implimented for the Microsoft
// Sidewinder Dual Strike gamepad under the Windows Operating System
//
// PINOCCHIO is complient with the LINUX operating system
//
// PINOCCHIO uses a joypad library 'Joypad.ccp' and 'Joypad.h' implimented by Rob Bateman 
// of the National Centre for Computer Craphics Bournemouth and i the programmer hereby 
// acknowledge all credit within this plugin code
//
//---------------------------------------------------------------------------------------
// PINOCCHIO version 1.1 Paul Lumley February 2005
//--------------------------------------------------------------
//INCLUDE FILES AND LIBRARIES

// required to link the libraries
#ifdef WIN32
	#pragma comment(lib,"Foundation.lib")
	#pragma comment(lib,"OpenMaya.lib")
	#pragma comment(lib,"OpenMayaFx.lib")
	#pragma comment(lib,"OpenMayaUi.lib")
	#pragma comment(lib,"Image.lib")
	#pragma comment(lib,"OpenMayaAnim.lib")
#endif

// under WIN32 we have to 'export' these functions so that they are visible
// inside the dll. If maya can't see the functions, it can't load the plugin!
#ifdef WIN32
	#define EXPORT __declspec(dllexport)
#else
	#define EXPORT
#endif
//--------------------------------------

#include <maya/MSimple.h>
#include <maya/MTime.h>
#include <maya/MGlobal.h>
#include <maya/MString.h>
#include <maya/MDagPath.h>
#include <maya/MFnDagNode.h>
#include <maya/MSelectionList.h>

//--------------------------------------
#include <maya/MFnPlugin.h>
#include <maya/MEventMessage.h>
#include <maya/MGlobal.h>
#include <maya/MString.h>
#include <maya/MPoint.h>
#include <maya/M3dView.h>
#include <time.h>
#include <stdio.h>
#include <fstream>
#include <string>
#include "Joypad.h"
//#include "FrameTimer.h"
//#include "pinocchio.h"

//--------------------------------------------------------------
//DEFINITIONS
//
// used to limit the axis coordinates between 1 and -1
#define JP_MAX_EXTENT 32766.0f
// dampening values
#define Y_LIMITER (g_pJoypad->Yaxis()/JP_MAX_EXTENT)
#define X_LIMITER (g_pJoypad->Xaxis()/JP_MAX_EXTENT)
// defines the joypad deadzone
#define DEAD 0.2
//--------------------------------------------------------------
//DECLARATIONS

using namespace std;

// pointer to the joypad
XInputJoypad* g_pJoypad = 0;

// axis struct has menbers: current and previous to hold information
// about the joypads axis coordinates for state functions

struct axis{
		float current;
		float previous;
	};
// button struct has members: state, change, current and previous to 
// hold information about buttons for state functions
struct button{
			int state;
			int change;
			int current;
			int previous;
	};

// the joypad struct has members: axis and buttons and represents 
// the phyisical device, also initialises some of its members

struct deviceJoypad{
		axis x, y;
		button A, B, C, D, X, Y, UP, L_TRIG, R_TRIG;
		deviceJoypad()
		{
			x.current=0.0;
			x.previous=0.0;
			y.current=0.0;
			y.previous=0.0;
			A.state=0;
			A.previous=0;
			B.state=0;
			B.previous=0;
			C.state=0;
			C.previous=0;
			D.state=0;
			D.previous=0;
			UP.state=0;
			UP.previous=0;
			X.state=0;
			X.previous=0;
			Y.state=0;
			Y.previous=0;
			L_TRIG.state=0;
			L_TRIG.previous=0;
			R_TRIG.state=0;
			R_TRIG.previous=0;
		}
	};
// the mayaObject struct has members: translate, rotate and misc and represents 
// the actual objects within maya
struct mayaObject{
		float translateX, translateY, translateZ;
		float rotateX, rotateY, rotateZ;
		int misc;
		mayaObject()
		{
			translateX=0.0;
			translateY=0.0;
			translateZ=0.0;
			rotateX=0.0;
			rotateY=0.0;
			rotateZ=0.0;
			misc=0;
			
		}
	};

	deviceJoypad joyPad;
	mayaObject rigCtrl, armCtrl, legCtrl;
	
//--------------------------------------------------------------
//MAYA CALLBACKS

// Having created a callback function within maya, we need to store the 
// generated MCallbackId so that we can delete the callback.

MCallbackId g_Callback;

//--------------------------------------------------------------
//JOYPAD FUNCTIONS
// button is() functions return true is a particular button is 
// being pressed these functions store the state information
// about the button

bool isA(){
		joyPad.A.state= g_pJoypad->Button(0);
		if(joyPad.A.state==1){
			joyPad.A.state=1;
			return TRUE;
		}
		else{
			joyPad.A.state=0;
			return FALSE;
		}
	}
bool isB(){
		joyPad.B.state = g_pJoypad->Button(1);
		if(joyPad.B.state==1){
			joyPad.B.state=1;
			return TRUE;
		}
		else{
			joyPad.B.state=0;
			return FALSE;
		}
	}
bool isC(){
		joyPad.C.state = g_pJoypad->Button(2);
		if(joyPad.C.state==1){
			joyPad.C.state=1;
			return TRUE;
		}
		else{
			joyPad.C.state=0;
			return FALSE;
		}
	}
bool isD(){
		joyPad.D.state = g_pJoypad->Button(3);
		if(joyPad.D.state==1){
			joyPad.D.state=1;
			return TRUE;
		}
		else{
			joyPad.D.state=0;
			return FALSE;
		}
	}
bool isUP(){
		joyPad.UP.state = g_pJoypad->Button(8);
		if(joyPad.UP.state==1){
			joyPad.UP.state=1;
			return TRUE;
		}
		else{
			joyPad.UP.state=0;
			return FALSE;
		}
	}
bool isX(){
		joyPad.X.state = g_pJoypad->Button(4);
		if(joyPad.X.state==1){
			joyPad.X.state=1;
			return TRUE;
		}
		else{
			joyPad.X.state=0;
			return FALSE;
		}
	}
bool isY(){
		joyPad.Y.state = g_pJoypad->Button(5);
		if(joyPad.Y.state==1){
			joyPad.Y.state=1;
			return TRUE;
		}
		else{
			joyPad.Y.state=0;
			return FALSE;
		}
	}
bool isL_TRIG(){
		joyPad.L_TRIG.state = g_pJoypad->Button(6);
		if(joyPad.L_TRIG.state==1){
			joyPad.L_TRIG.state=1;
			return TRUE;
		}
		else{
			joyPad.L_TRIG.state=0;
			return FALSE;
		}
	}
bool isR_TRIG(){
		joyPad.R_TRIG.state = g_pJoypad->Button(7);
		if(joyPad.R_TRIG.state==1){
			joyPad.R_TRIG.state=1;
			return TRUE;
		}
		else{
			joyPad.R_TRIG.state=0;
			return FALSE;
		}
	}

//--------------------------------------------------------------
// isbutton() function returns true if any one of the joypads
// buttons is being pressed

int isButton(void){

	if(isA()||isB()||isC()||isD()||isX()||isY()||isUP()||isL_TRIG()||isR_TRIG())
		return 1;
	else
		return 0;

}


//--------------------------------------------------------------
// hasChanged() functions return true if the state of the 
// button has changed used to attach functions to a button press

int hasChanged_A(){
	int retVal;

	joyPad.A.current=g_pJoypad->Button(0);

	if(joyPad.A.previous-joyPad.A.current==1)
		retVal = 1;
	else
		retVal = 0;

	joyPad.A.previous = joyPad.A.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_B(){
	int retVal;

	joyPad.B.current=g_pJoypad->Button(1);

	if(joyPad.B.previous-joyPad.B.current==1)
		retVal = 1;
	else
		retVal = 0;
	
	joyPad.B.previous = joyPad.B.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_C(){
	int retVal;
	
	joyPad.C.current=g_pJoypad->Button(2);

	if(joyPad.C.previous-joyPad.C.current==1)
		retVal = 1;
	else
		retVal = 0;
	
	joyPad.C.previous = joyPad.C.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_D(){
	int retVal;

	joyPad.D.current=g_pJoypad->Button(3);

	if(joyPad.D.previous-joyPad.D.current==1)
		retVal = 1;
	else
		retVal = 0;
	
	joyPad.D.previous = joyPad.D.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_UP(){
	int retVal;
	
	joyPad.UP.current=g_pJoypad->Button(8);

	if(joyPad.UP.previous-joyPad.UP.current==1)
		retVal = 1;
	else
		retVal = 0;
	
	joyPad.UP.previous = joyPad.UP.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_X(){
	int retVal;
	
	joyPad.X.current=g_pJoypad->Button(4);

	if(joyPad.X.previous-joyPad.X.current==1)
		retVal = 1;
	else
		retVal = 0;

	joyPad.X.previous = joyPad.X.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_Y(){
	int retVal;

	joyPad.Y.current=g_pJoypad->Button(5);

	if(joyPad.Y.previous-joyPad.Y.current==1)
		retVal = 1;
	else
		retVal = 0;

	joyPad.Y.previous = joyPad.Y.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_L_TRIG(){
	int retVal;
	
	joyPad.L_TRIG.current=g_pJoypad->Button(6);

	if(joyPad.L_TRIG.previous-joyPad.L_TRIG.current==1)
		retVal = 1;
	else
		retVal = 0;
	
	joyPad.L_TRIG.previous = joyPad.L_TRIG.current;
	UpdateInputState();
	return retVal;
}
int hasChanged_R_TRIG(){
	int retVal;

	joyPad.R_TRIG.current=g_pJoypad->Button(7);

	if(joyPad.R_TRIG.previous-joyPad.R_TRIG.current==1)
		retVal = 1;
	else
		retVal = 0;

	joyPad.R_TRIG.previous = joyPad.R_TRIG.current;
	UpdateInputState();
	return retVal;
}
//--------------------------------------------------------------
// axisStateChange() functions return true or false if the axis
// is being moved

int xAxisStateChange(){	

	
	int retVal;
	joyPad.x.current=g_pJoypad->Xaxis()/JP_MAX_EXTENT;

		if(joyPad.x.previous-joyPad.x.current!=0){
			retVal = 1;
		}
		else{
			retVal = 0;
		}
	joyPad.x.previous = joyPad.x.current;
	UpdateInputState();
	return retVal;
}
int yAxisStateChange(){	
			
			int retVal;

				joyPad.y.current = g_pJoypad->Yaxis()/JP_MAX_EXTENT;

				if(joyPad.y.previous-joyPad.y.current!=0){
					retVal = 1;
				}
				else{
					retVal = 0;
				}
				joyPad.y.previous = joyPad.y.current;
				UpdateInputState();
				return retVal;
}


// axisState() functions return the curent position of the axis-
// up, down, left, right or centre

int xAxisState(){	
						
			int retVal; float tempVal;

			tempVal = g_pJoypad->Xaxis()/JP_MAX_EXTENT;
			if(tempVal>-DEAD&&tempVal<DEAD)
				retVal=2;//centre

			else if(tempVal>0.2)
				retVal=0;//right
				
			else
				retVal=1;//left

			UpdateInputState();
			return retVal;
}
int yAxisState(){	
						
			int retVal; float tempVal;

			tempVal = g_pJoypad->Yaxis()/JP_MAX_EXTENT;
			if(tempVal>-DEAD&&tempVal<DEAD)
				retVal=2;//centre

			else if(tempVal>0.2)
				retVal=1;//up
				
			else
				retVal=0;//down

			UpdateInputState();
			return retVal;
}

//--------------------------------------------------------------
//API FUNCTIONS()
//
void setTranslationSelection(float _tx, float _ty, float _tz, int selection){

    MDagPath            node;
    MObject             component;
    MSelectionList      list;
    MFnDagNode          nodeFn;
	MPlug				plug;

    MGlobal::getActiveSelectionList( list );

        list.getDagPath( selection, node, component );
        nodeFn.setObject( node );
		plug = nodeFn.findPlug("translate");

			if(plug.isCompound()) {
  	
				// get the plugs to the child attributes
				MPlug plug_x = plug.child(0);
				MPlug plug_y = plug.child(1);
				MPlug plug_z = plug.child(2);

				// get the values from the child plugs
				plug_x.setValue( _tx );
				plug_y.setValue( _ty );
				plug_z.setValue( _tz );

			}
}
void setRotationSelection(float _rx, float _ry, float _rz, int selection){

    MDagPath            node;
    MObject             component;
    MSelectionList      list;
    MFnDagNode          nodeFn;
	MPlug				plug;

    MGlobal::getActiveSelectionList( list );

        list.getDagPath( selection, node, component );
        nodeFn.setObject( node );
		plug = nodeFn.findPlug("rotate");

			if(plug.isCompound()) {
  	
				// get the plugs to the child attributes
				MPlug plug_x = plug.child(0);
				MPlug plug_y = plug.child(1);
				MPlug plug_z = plug.child(2);

				// get the values from the child plugs
				plug_x.setValue( _rx );
				plug_y.setValue( _ry );
				plug_z.setValue( _rz );

			}
}
void setArmIKinematicsSelection(int setIK, int selection){

    MDagPath            node;
    MObject             component;
    MSelectionList      list;
    MFnDagNode          nodeFn;
	MPlug				plug;

    MGlobal::getActiveSelectionList( list );

        list.getDagPath( selection, node, component );
        nodeFn.setObject( node );
		plug = nodeFn.findPlug("armIK");

			// get the values from the plugs
			plug.setValue( setIK );
}
//--------------------------------------------------------------
//IDLE FUNCTION()
//
void IdleFunc(void* ){

	MTime time;

	MGlobal::selectByName( "marionetteCtrl", MGlobal::kAddToList);
	MGlobal::selectByName( "control_arms", MGlobal::kAddToList);
	MGlobal::selectByName( "control_legs", MGlobal::kAddToList);

	if(hasChanged_L_TRIG()){//reset puppet

		rigCtrl.rotateX=0;
		rigCtrl.rotateY=0;
		rigCtrl.rotateZ=0;
		rigCtrl.translateX=0;
		rigCtrl.translateY=0;
		rigCtrl.translateZ=0;

		armCtrl.translateX=0;
		armCtrl.translateY=0;
		armCtrl.translateZ=0;
		armCtrl.rotateX=0;
		armCtrl.rotateY=0;
		armCtrl.rotateZ=0;
		armCtrl.misc=1;

		legCtrl.translateX=0;
		legCtrl.translateY=0;
		legCtrl.translateZ=0;
		legCtrl.rotateX=0;
		legCtrl.rotateY=0;
		legCtrl.rotateZ=0;
	}

		//------------------------------------------------------------
		//RIG MOVEMENT(selection index 0)

	if(!isX()&&!isY()){// X and Y are not being pressed
		//if X axis then bend LEFT RIGHT

			if(xAxisState()==2)//centre
				rigCtrl.rotateZ += 0.0;
			if(xAxisState()==1)//left
				rigCtrl.rotateZ += 0.05f*X_LIMITER;
			if(xAxisState()==0)//right
				rigCtrl.rotateZ += 0.05f*X_LIMITER;

		//if Y axis then bend FORWARD BACK

			if(yAxisState()==2)//centre
				rigCtrl.rotateX += 0.0;
			if(yAxisState()==1)//up
				rigCtrl.rotateX -= 0.05f*Y_LIMITER;
			if(yAxisState()==0)//down
				rigCtrl.rotateX -= 0.05f*Y_LIMITER;

		//if button A then move rig FORWARD

			if(isA()&&!isUP())
				rigCtrl.translateZ += 1;

		//if button D then move rig BACKWARD

			if(isD()&&!isUP())
				rigCtrl.translateZ -= 1;

		//if button B then move rig LEFT

			if(isB())
				rigCtrl.translateX += 1;

		//if button C then move rig RIGHT

			if(isC())
				rigCtrl.translateX -= 1;

		//if button UP && A then move rig UP

			if(isUP()&&isA())
				rigCtrl.translateY += 1;

		//if button UP && D then move rig DOWN
			if(isUP()&&isD())
				rigCtrl.translateY -= 1;
	}
		//------------------------------------------------------------
		//------------------------------------------------------------
		//ARM MOVEMENT(selection index 1)

		//if button X && X axis then SWING arms 

			if(isX()&&xAxisState()==2)//centre
				armCtrl.rotateZ += 0.0;
			if(isX()&&xAxisState()==1)//left
				armCtrl.rotateZ -= 0.5f*X_LIMITER;
			if(isX()&&xAxisState()==0)//right
				armCtrl.rotateZ -= 0.5f*X_LIMITER;

		//if button X && Y axis then move arms UP/DOWN

			if(isX()&&yAxisState()==2)//centre
				armCtrl.translateY += 0.0;
			if(isX()&&yAxisState()==1)//up
				armCtrl.translateY -= 10*Y_LIMITER;
			if(isX()&&yAxisState()==0)//down
				armCtrl.translateY -= 10*Y_LIMITER;

		//if button X && button A then move arms FORWARD

			if(isX()&&isA())
				armCtrl.translateZ += 1;

		//if button X && button D then move arms BACKWARD

			if(isX()&&isD())
				armCtrl.translateZ -= 1;

		//if button X && button B then move arms LEFT

			if(isX()&&isB())
				armCtrl.translateX += 1;

		//if button X && button C then move arms RIGHT

			if(isX()&&isC())
				armCtrl.translateX -= 1;

		//if button X && button UP(change) then SWITCH KINEMATICS

			if(isX()&&hasChanged_UP()){
				if(armCtrl.misc==1)
					armCtrl.misc=0;
				else
					armCtrl.misc=1;
			}

		//if button X && button R_TRIG then perform marionette translation controls

		//if button A then move rig FORWARD

			if(isA()&&isX()&&isR_TRIG())
				rigCtrl.translateZ += 2;

		//if button D then move rig BACKWARD

			if(isD()&&isX()&&isR_TRIG())
				rigCtrl.translateZ -= 1;

		//if button B then move rig LEFT

			if(isB()&&isX()&&isR_TRIG())
				rigCtrl.translateX += 1;

		//if button C then move rig RIGHT

			if(isC()&&isX()&&isR_TRIG())
				rigCtrl.translateX -= 1;

		//------------------------------------------------------------
		//------------------------------------------------------------
		//LEG MOVEMENT(selection index 2)

		//if button Y && X axis then SWING legs

			if(isY()&&xAxisState()==2)//centre
				legCtrl.rotateY += 0.0;
			if(isY()&&xAxisState()==1)//left
				legCtrl.rotateY -= 0.05f*X_LIMITER;
			if(isY()&&xAxisState()==0)//right
				legCtrl.rotateY -= 0.05f*X_LIMITER;

		//if button Y && Y axis then RAISE/LOWER legs

			if(isY()&&yAxisState()==2)//centre
				legCtrl.rotateZ += 0.0;
			if(isY()&&yAxisState()==1)//up
				legCtrl.rotateZ -= 0.5f*Y_LIMITER;
			if(isY()&&yAxisState()==0)//down
				legCtrl.rotateZ -= 0.5f*Y_LIMITER;

		//if button Y && button A then move legs FORWARD

			if(isY()&&isA()&&!isUP())
				legCtrl.translateZ += 1;

		//if button Y && button D then move legs BACKWARD

			if(isY()&&isD()&&!isUP())
				legCtrl.translateZ -= 1;

		//if button Y && button B then move legs LEFT

			if(isY()&&isB())
				legCtrl.translateX += 1;

		//if button Y && button C then move legs RIGHT

			if(isY()&&isC())
				legCtrl.translateX -= 1;

		//if button Y && button UP && A then move legs UP

			if(isUP()&&isY()&&isA())
				legCtrl.translateY += 1;

		//if button Y && button UP && D then move legs DOWN

			if(isUP()&&isY()&&isD())
				legCtrl.translateY -= 1;

		//if button Y && button R_TRIG then perform marionette translation controls

		//if button A then move rig FORWARD

			if(isA()&&isY()&&isR_TRIG())
				rigCtrl.translateZ += 1;

		//if button D then move rig BACKWARD

			if(isD()&&isY()&&isR_TRIG())
				rigCtrl.translateZ -= 1;

		//if button B then move rig LEFT

			if(isB()&&isY()&&isR_TRIG())
				rigCtrl.translateX += 1;

		//if button C then move rig RIGHT

			if(isC()&&isY()&&isR_TRIG())
				rigCtrl.translateX -= 1;
		//------------------------------------------------------------

 			setTranslationSelection(rigCtrl.translateX, rigCtrl.translateY, rigCtrl.translateZ, 0);
			setRotationSelection(rigCtrl.rotateX, rigCtrl.rotateY, rigCtrl.rotateZ, 0);

			setTranslationSelection(armCtrl.translateX, armCtrl.translateY, armCtrl.translateZ, 1);
			setRotationSelection(armCtrl.rotateX, armCtrl.rotateY, armCtrl.rotateZ, 1);
			setArmIKinematicsSelection(armCtrl.misc, 1);

			setTranslationSelection(legCtrl.translateX, legCtrl.translateY, legCtrl.translateZ, 2);
			setRotationSelection(legCtrl.rotateX, legCtrl.rotateY, legCtrl.rotateZ, 2);

			

}
//--------------------------------------------------------------
//API PLUGIN FUNCTIONS

// INITIALIZEPLUGIN()
// This function is called once when our plugin is loaded. We can use
// it to register any custom nodes, mel functions etc.

EXPORT MStatus initializePlugin( MObject obj )
{
	// we need to use the plugin function set to register a new mel command with maya.
	// we can provide some info about the author of the plugin, the version number, and we
	// can also specify a required version of maya, though usually "Any" will do...
	
	MFnPlugin fnPlugin( obj, "Paul Lumley", "1.0", "Any");


	// only create callback if input updated OK
	if( InitInput() && NumJoypads() )
	{
		// get a pointer to the joypad
		g_pJoypad = GetJoypad(0);

		/// When creating the callback for the idle function, we need to specify a callback
		/// type. In this case we are using "idle", however it can be anyone of the strings
		/// that `scriptJob -listEvents` returns;
		/// The second parameter is a pointer to the callback function. The third is a user
		/// data parameter (which is explained in MEventMessage2 example). In this case
		/// we can use NULL since we dont want any user data
		///
		g_Callback = MEventMessage::addEventCallback( "idle", IdleFunc, NULL );

	}
	else {
		std::cerr << "Could not initialise joypads\n";
		return MS::kFailure;
	}

	return MS::kSuccess;
}

// UNINITIALIZEPLUGIN()
// This function is called once when our plugin is unloaded. We need
// to tell maya which mel funcs, nodes etc we are removing.

EXPORT MStatus uninitializePlugin( MObject obj)
{
	// use the plugin function set to unload the plugin
	MFnPlugin fnPlugin( obj );

	if(g_pJoypad) {
		/// make sure we delete the callback before the plugin quits !!!
		/// Maya is likely to crash otherwise!
		MEventMessage::removeCallback( g_Callback );
	}

	/// release DX8 or joystick handles for linux
	FreeInput();
	return MS::kSuccess;
}